@@ -3,8 +3,9 @@ __submodules__ = ["core"]
 
 # <AUTOGEN_INIT>
 from snakebids.core import (
-    BidsInputs,
-    BidsInputsDict,
+    BidsComponent,
+    BidsDataset,
+    BidsDatasetDict,
     bids,
     filter_list,
     generate_inputs,
@@ -15,8 +16,9 @@ from snakebids.core import (
 )
 
 __all__ = [
-    "BidsInputs",
-    "BidsInputsDict",
+    "BidsComponent",
+    "BidsDataset",
+    "BidsDatasetDict",
     "bids",
     "filter_list",
     "generate_inputs",
