@@ -5,65 +5,40 @@ import json
 import logging
 import operator as op
 import re
-from collections import defaultdict
+from collections import UserDict, defaultdict
 from pathlib import Path
+from string import Formatter
 from typing import (
+    TYPE_CHECKING,
+    Any,
     Dict,
     Generator,
     Iterable,
     List,
-    NamedTuple,
     Optional,
     Tuple,
     Union,
+    cast,
     overload,
 )
 
 import attr
 import more_itertools as itx
 from bids import BIDSLayout, BIDSLayoutIndexer
+from bids.layout import Query
+from cached_property import cached_property
 from typing_extensions import Literal, TypedDict
 
+import snakebids.utils.sb_itertools as sb_it
 from snakebids.core.filtering import filter_list
+from snakebids.exceptions import ConfigError
 from snakebids.utils.snakemake_io import glob_wildcards
-from snakebids.utils.utils import get_match_search_func, read_bids_tags
+from snakebids.utils.utils import BidsEntity, property_alias
 
 _logger = logging.getLogger(__name__)
 
 
-_BidsLists = NamedTuple(
-    "BidsLists",
-    [
-        ("input_name", str),
-        ("input_path", str),
-        ("input_zip_lists", Dict[str, List[str]]),
-        ("input_lists", Dict[str, List[str]]),
-        ("input_wildcards", Dict[str, str]),
-    ],
-)
-_BidsLists.__doc__ = """Bids inputs from a single input descriptor
-
-Attributes
-----------
-input_name : str
-    The name of the input descriptor.
-input_path : str
-    Wildcard-filled path that matches the images for this modality.
-input_zip_lists : dict
-    Each key is a wildcard entity and each value is a list of the values found for that
-    entity. Each of these lists has length equal to the number of images matched for
-    this modality, so they can be zipped together to get a list of the wildcard values
-    for each image.
-input_lists : dict
-    Each key is a wildcard entity and each value is a list of the unique values found
-    for that entity. These lists might not be the same length.
-input_wildcards : dict
-    Where each key is the name of a wildcard entity, and each value is the Snakemake
-    wildcard used for that entity.
-"""
-
-
-class BidsInputsDict(TypedDict):
+class BidsDatasetDict(TypedDict):
     """Dict equivalent of BidsInputs, for backwards-compatibility"""
 
     input_path: Dict[str, str]
@@ -75,59 +50,251 @@ class BidsInputsDict(TypedDict):
     subj_wildcards: Dict[str, str]
 
 
-@attr.frozen
-class BidsInputs:
-    """Bids input entities and their resolved values
+@attr.define
+class BidsComponent:
+    """Component of a BidsDataset mapping entities to their resolved values
 
-    Attributes
+    BidsComponents are immutable: their values cannot be altered.
+
+    Properties
     ----------
-    input_path
-        Wildcard-filled path that matches the images for this modality.
-    input_zip_lists
+    name
+        Name of the component
+    path
+        Wildcard-filled path that matches the files for this component.
+    zip_lists
         Dictionary where each key is a wildcard entity and each value is a list of the
         values found for that entity. Each of these lists has length equal to the number
         of images matched for this modality, so they can be zipped together to get a
-        list of the wildcard values for each image.
-    input_lists
-        Dictionary where each key is a wildcard entity and each value is a list of the
-        unique values found for that entity. These lists might not be the same length.
-    input_wildcards
-        Dictionary where each key is the name of a wildcard entity, and each value is
-        the Snakemake wildcard used for that entity.
+        list of the wildcard values for each file.
     """
 
-    input_path: Dict[str, str]
-    input_zip_lists: Dict[str, Dict[str, List[str]]]
-    input_lists: Dict[str, Dict[str, List[str]]]
-    input_wildcards: Dict[str, Dict[str, str]]
+    name: str = attr.field(on_setattr=attr.setters.frozen)
+    path: str = attr.field(on_setattr=attr.setters.frozen)
+    zip_lists: Dict[str, List[str]] = attr.field(
+        on_setattr=attr.setters.frozen, converter=dict
+    )
+
+    @zip_lists.validator  # type: ignore
+    def _validate_zip_lists(self, _, value: Dict[str, List[str]]):
+        lengths = {len(val) for val in value.values()}
+        if len(lengths) > 1:
+            raise ValueError("zip_lists must all be of equal length")
+        _, raw_fields, *_ = sb_it.unpack(
+            zip(*Formatter().parse(self.path)), [[], [], []]
+        )
+        fields = set(filter(None, raw_fields))
+        if fields != set(value):
+            raise ValueError(
+                "zip_lists entries must match the wildcards in input_path: "
+                f"Path: {fields} != zip_lists: {set(value)}"
+            )
 
     @property
+    def input_name(self):
+        """Alias of :attr:`name <snakebids.BidsComponent.name>`
+
+        Name of the component
+        """
+        return self.name
+
+    @property
+    def input_path(self):
+        """Alias of :attr:`path <snakebids.BidsComponent.path>`
+
+        Wildcard-filled path that matches the files for this component.
+        """
+        return self.path
+
+    @property
+    def input_zip_lists(self):
+        """Alias of :attr:`zip_lists <snakebids.BidsComponent.zip_lists>`
+
+        Dictionary where each key is a wildcard entity and each value is a list of the
+        values found for that entity. Each of these lists has length equal to the number
+        of images matched for this modality, so they can be zipped together to get a
+        list of the wildcard values for each file.
+        """
+        return self.zip_lists
+
+    # Note: we can't use cached property here because it's incompatible with slots.
+    _input_lists: Optional[Dict[str, List[str]]] = attr.field(
+        default=None, init=False, eq=False, repr=False
+    )
+    _input_wildcards: Optional[Dict[str, str]] = attr.field(
+        default=None, init=False, eq=False, repr=False
+    )
+    _entities: Optional[List[str]] = attr.field(
+        default=None, init=False, eq=False, repr=False
+    )
+
+    @property
+    def entities(self):
+        """Component entities and their associated values
+
+        Dictionary where each key is an entity and each value is a list of the
+        unique values found for that entity. These lists might not be the same length.
+        """
+        if self._input_lists is None:
+            self._input_lists = {
+                entity: list(set(values)) for entity, values in self.zip_lists.items()
+            }
+        return self._input_lists
+
+    @property_alias(entities, "entities", "snakebids.BidsComponent.entities")
+    def input_lists(self):
+        return self.entities
+
+    @property
+    def wildcards(self):
+        """Wildcards in brace-wrapped syntax
+
+        Dictionary where each key is the name of a wildcard entity, and each value is
+        the Snakemake wildcard used for that entity.
+        """
+        if self._input_wildcards is None:
+            self._input_wildcards = {
+                entity: f"{{{entity}}}" for entity in self.zip_lists
+            }
+        return self._input_wildcards
+
+    @property_alias(wildcards, "wildcards", "snakebids.BidsComponent.wildcards")
+    def input_wildcards(self):
+        return self.wildcards
+
+    def __eq__(self, other: Union["BidsComponent", object]):
+        if not isinstance(other, BidsComponent):
+            return False
+
+        if self.name != other.name:
+            return False
+
+        if self.path != other.path:
+            return False
+
+        def sorted_items(dictionary: Dict[str, List[str]]):
+            return sorted(dictionary.items(), key=op.itemgetter(0))
+
+        if set(self.zip_lists) != set(other.zip_lists):
+            return False
+
+        if not other.zip_lists and not self.zip_lists:
+            return True
+
+        other_items = cast(
+            List[List[str]], list(zip(*sorted_items(other.zip_lists)))[1]
+        )
+        our_items = cast(List[List[str]], list(zip(*sorted_items(self.zip_lists)))[1])
+
+        return set(zip(*our_items)) == set(zip(*other_items))
+
+
+if TYPE_CHECKING:
+    _BidsComponentsType = UserDict[str, BidsComponent]
+else:
+    # UserDict is not subscriptable in py37
+    _BidsComponentsType = UserDict
+
+
+class BidsDataset(_BidsComponentsType):
+    """A bids dataset parsed by pybids, organized into BidsComponents.
+
+    BidsDatasets are typically generated using :func:`generate_inputs()
+    <snakebids.generate_inputs>`, which reads the ``pybids_inputs`` field in your
+    snakemake config file and, for each entry, creates a BidsComponent using the
+    provided name, wildcards, and filters.
+
+    Individual components can be accessed using bracket-syntax: (e.g.
+    ``inputs["t1w"]``). Component access attributes along with the component name in
+    brackets can also be used. For example, ``BidsDataset.entities["t1w"]`` and
+    ``BidsDataset["t1w"].entities`` return the same thing.
+
+    Provides access to summarizing information, for instance, the set of all subjects or
+    sessions found in the dataset
+    """
+
+    # pylint: disable=super-init-not-called
+    def __init__(self, data: Any):
+        self.data = dict(data)
+
+    def __setitem__(self, _: Any, __: Any):
+        raise NotImplementedError(
+            f"Modification of {self.__class__.__name__} is not yet supported"
+        )
+
+    @cached_property
+    def path(self):
+        """Dict mapping :class:`BidsComponents <snakebids.BidsComponent>` names to \
+        their ``paths``.
+        """
+        return {key: value.path for key, value in self.data.items()}
+
+    @property_alias(path, "path", "snakebids.BidsDataset.path")
+    def input_path(self):
+        return self.path
+
+    @cached_property
+    def zip_lists(self):
+        """Dict mapping :class:`BidsComponents <snakebids.BidsComponent>` names to \
+        their ``zip_lists``
+        """
+        return {key: value.zip_lists for key, value in self.data.items()}
+
+    @property_alias(zip_lists, "zip_lists", "snakebids.BidsDataset.zip_lists")
+    def input_zip_lists(self):
+        return self.zip_lists
+
+    @cached_property
+    def entities(self):
+        """Dict mapping :class:`BidsComponents <snakebids.BidsComponent>` names to \
+        to their :attr:`entities <snakebids.BidsComponent.entities>`
+        """
+        return {key: value.entities for key, value in self.data.items()}
+
+    @property_alias(entities, "entities", "snakebids.BidsDataset.entities")
+    def input_lists(self):
+        return self.entities
+
+    @cached_property
+    def wildcards(self):
+        """Dict mapping :class:`BidsComponents <snakebids.BidsComponent>` names to \
+        their :attr:`wildcards <snakebids.BidsComponent.wildcards>`
+        """
+        return {key: value.input_wildcards for key, value in self.data.items()}
+
+    @property_alias(wildcards, "wildcards", "snakebids.BidsDataset.wildcards")
+    def input_wildcards(self):
+        return self.wildcards
+
+    @cached_property
     def subjects(self):
         """A list of the subjects in the dataset."""
         return [
             *{
                 *it.chain.from_iterable(
                     input_list["subject"]
-                    for input_list in self.input_lists.values()
+                    for input_list in self.entities.values()
                     if "subject" in input_list
                 )
             }
         ]
 
     @property
+    # @ft.lru_cache(None)
     def sessions(self):
         """A list of the sessions in the dataset."""
         return [
             *{
                 *it.chain.from_iterable(
                     input_list["session"]
-                    for input_list in self.input_lists.values()
+                    for input_list in self.entities.values()
                     if "session" in input_list
                 )
             }
         ]
 
     @property
+    # @ft.lru_cache(None)
     def subj_wildcards(self):
         """The subject and session wildcards applicable to this dataset.
 
@@ -143,60 +310,38 @@ class BidsInputs:
 
     @property
     def as_dict(self):
-        return BidsInputsDict(
-            **attr.asdict(self),
+        """Get the layout as a legacy dict
+
+        Included primarily for backward compatability with older versions of snakebids,
+        where generate_inputs() returned a dict rather than the `BidsDataset` class
+
+        Returns
+        -------
+        BidsDatasetDict
+        """
+        return BidsDatasetDict(
+            input_path=self.input_path,
+            input_lists=self.entities,
+            input_wildcards=self.input_wildcards,
+            input_zip_lists=self.input_zip_lists,
             subjects=self.subjects,
             sessions=self.sessions,
             subj_wildcards=self.subj_wildcards,
         )
 
-    @staticmethod
-    def _get_bids_lists_as_dicts(bids_lists: Iterable[_BidsLists]):
-        """Convert Iterable of Bids lists into dictionaries
-
-        The dictionaries will be in a form suitable for use in BidsInputs. In other
-        words, each list type (input_path, input_zip_lists, etc) will be a dict with
-        input descriptors as keys and the list type as values.
-
-        Parameters
-        ----------
-        bids_lists : Iterable[BidsLists]
-            BidsLists to be converted
-
-        Returns
-        -------
-        Tuple of Dicts
-            Tuple of all the dicts generated from the BidsLists
-        """
-        bids_lists = [*bids_lists]
-
-        def get_as_dict(attr_name):
-            return {
-                bids_list.input_name: getattr(bids_list, attr_name)
-                for bids_list in bids_lists
-            }
-
-        return (
-            get_as_dict("input_path"),
-            get_as_dict("input_zip_lists"),
-            get_as_dict("input_lists"),
-            get_as_dict("input_wildcards"),
-        )
-
     @classmethod
-    def from_bids_lists(cls, bids_lists: Iterable[_BidsLists]):
-        """Generate BidsInputs based on an iterable of BidsLists
+    def from_iterable(cls, iterable: Iterable[BidsComponent]):
+        """Construct Dataset from iterable of BidsComponents
 
         Parameters
         ----------
-        bids_lists : Iterable[BidsLists]
-            Iterable of BidsLists to convert
+        iterable : Iterable[BidsComponent]
 
         Returns
         -------
-        BidsInputs
+        BidsDataset
         """
-        return cls(*cls._get_bids_lists_as_dicts(bids_lists))
+        return cls({bidsinput.name: bidsinput for bidsinput in iterable})
 
 
 # pylint: disable=too-many-arguments
@@ -212,7 +357,7 @@ def generate_inputs(
     participant_label=...,
     exclude_participant_label=...,
     use_bids_inputs: Union[Literal[False], None] = ...,
-) -> BidsInputsDict:
+) -> BidsDatasetDict:
     ...
 
 
@@ -229,7 +374,7 @@ def generate_inputs(
     participant_label=...,
     exclude_participant_label=...,
     use_bids_inputs: Literal[True] = ...,
-) -> BidsInputs:
+) -> BidsDataset:
     ...
 
 
@@ -314,7 +459,7 @@ def generate_inputs(
 
     Returns
     -------
-    BidsInputs or BidsInputsDict:
+    BidsDataset or BidsDatasetDict:
         Object containing organized information about the bids inputs for consumption
         in snakemake. See the documentation of BidsInputs for details and examples.
 
@@ -402,23 +547,23 @@ def generate_inputs(
     Or, if BidsInputs is enabled::
 
         <class BidsInputs>
-            input_path: {
+            path: {
                 "bold": "bids-example/sub-{subject}/func/sub-{subject}_task-{task}_bold\
                     .nii.gz"
             }
-            input_zip_lists: {
+            zip_lists: {
                 "bold": {
                     "subject": ["control01", "control01"],
                     "task": ["nback", "rest"]
                 }
             }
-            input_lists: {
+            entities: {
                 "bold": {
                     "subject": ["control01"],
                     "task": ["nback", "rest"]
                 }
             }
-            input_wildcards: {
+            wildcards: {
                 "bold": {
                     "subject": "{subject}",
                     "task": "{task}"
@@ -442,8 +587,6 @@ def generate_inputs(
         pybids_reset_database=pybids_reset_database,
     )
 
-    # this will populate input_path, input_lists, input_zip_lists, and
-    # input_wildcards
     filters = {"subject": subject_filter} if subject_filter else {}
     bids_inputs = _get_lists_from_bids(
         bids_layout=layout,
@@ -468,8 +611,8 @@ def generate_inputs(
         )
         use_bids_inputs = False
     if use_bids_inputs:
-        return BidsInputs.from_bids_lists(bids_inputs)
-    return BidsInputs.from_bids_lists(bids_inputs).as_dict
+        return BidsDataset.from_iterable(bids_inputs)
+    return BidsDataset.from_iterable(bids_inputs).as_dict
 
 
 def _gen_bids_layout(
@@ -659,48 +802,22 @@ def _parse_custom_path(
         _logger.warning("No wildcards defined in %s", input_path)
 
     # Initialize output values
-    input_wildcards: Dict[str, str] = {}
-    input_zip_lists: Dict[str, List[str]] = {}
-    input_lists: Dict[str, List[str]] = {}
+    zip_lists: Dict[str, List[str]] = {}
 
     # Log an error if no matches found
-    # TODO: This will fail to detect filtering correctly as, up till now, it has
-    #       only been performed on input_lists
     if len(wildcards[0]) == 0:
         _logger.error("No matching files for %s", input_path)
-        return (input_zip_lists, input_lists, input_wildcards)
+        return zip_lists
 
     # Loop through every wildcard name
     for i, wildcard in enumerate(wildcard_names):
-        # Check if this wildcard needs to be filtered
-        if wildcard not in filters:
-            # If not, the match_func will always return True
-            match_func = lambda _: True  # noqa: E731
-        elif regex_search:
-            # Otherwise, we use regex matching for regex_search
-            match_func = get_match_search_func(
-                itx.always_iterable(filters[wildcard]), re.match
-            )
-        else:
-            # And a simple equality operator for everything else
-            match_func = get_match_search_func(
-                itx.always_iterable(filters[wildcard]), op.eq
-            )
+        zip_lists[wildcard] = wildcards[i]
 
-        # Add the wildcard item to each output value, using filtering for input_lists
-        input_zip_lists[wildcard] = wildcards[i]
-        input_lists[wildcard] = [*filter(match_func, set(wildcards[i]))]
-        input_wildcards[wildcard] = f"{{{wildcard}}}"
-
-    # Return the output values, running filtering on the input_zip_lists
-    return (
-        filter_list(input_zip_lists, filters, regex_search=regex_search),
-        input_lists,
-        input_wildcards,
-    )
+    # Return the output values, running filtering on the zip_lists
+    return filter_list(zip_lists, filters, regex_search=regex_search)
 
 
-def _parse_bids_path(path: str, wildcards: Iterable[str]) -> Tuple[str, Dict[str, str]]:
+def _parse_bids_path(path: str, entities: Iterable[str]) -> Tuple[str, Dict[str, str]]:
     """Replace parameters in an bids path with the given wildcard {tags}.
 
     Parameters
@@ -721,43 +838,34 @@ def _parse_bids_path(path: str, wildcards: Iterable[str]) -> Tuple[str, Dict[str
     """
 
     wildcard_values: Dict[str, str] = {}
-    bids_tags = read_bids_tags()
 
-    for wildcard in wildcards:
+    # BUG: No support for "datatype"
+    for entity in map(BidsEntity, entities):
         # Iterate over wildcards, slowly updating the path as each entity is replaced
 
-        tag = bids_tags[wildcard] if wildcard in bids_tags else wildcard
+        tag = entity.tag
+        wildcard = entity.wildcard
 
-        # this changes e.g. sub-001 to sub-{subject} in the path
-        # (so snakemake can use the wildcards)
-        # HACK FIX FOR acq vs acquisition etc -- should
-        # eventually update the bids() function to also use
-        # bids_tags.json, where e.g. acquisition -> acq is
-        # defined.. -- then, can use wildcard_name instead
-        # of out_name..
-        if wildcard not in ["subject", "session"]:
-            out_name = tag
-        else:
-            out_name = wildcard
-
-        if wildcard == "suffix":
+        if entity == "suffix":
             # capture suffix
             match = re.search(r".*_([a-zA-Z0-9]+).*$", path)
 
             # capture "(before)suffix(after)" and replace with "before{suffix}after"
-            new_path = re.sub(r"(.*_)[a-zA-Z0-9]+(.*)$", rf"\1{{{out_name}}}\2", path)
+            new_path = re.sub(r"(.*_)[a-zA-Z0-9]+(.*)$", rf"\1{{{wildcard}}}\2", path)
 
         else:
+            # BUG: tags cannot be a substr of another tag in the same path (e.g. space
+            #      and ce)
             pattern = f"{tag}-([a-zA-Z0-9]+)"
-            replace = f"{tag}-{{{out_name}}}"
+            replace = f"{tag}-{{{wildcard}}}"
 
             match = re.search(pattern, path)
             new_path = re.sub(pattern, replace, path)
 
         if match and match.group(1):
-            entity = match.group(1)
+            value = match.group(1)
         else:
-            entity = ""
+            value = ""
 
         # update the path with the {wildcards} -- uses the
         # value from the string (not from the pybids
@@ -765,14 +873,18 @@ def _parse_bids_path(path: str, wildcards: Iterable[str]) -> Tuple[str, Dict[str
         # formatting (e.g. for run=01)
 
         path = new_path
-        wildcard_values[out_name] = entity
+        wildcard_values[wildcard] = value
 
     return path, wildcard_values
 
 
 def _get_lists_from_bids(
-    bids_layout: Optional[BIDSLayout], pybids_inputs, limit_to=None, **filters
-) -> Generator[_BidsLists, None, None]:
+    bids_layout: Optional[BIDSLayout],
+    pybids_inputs,
+    *,
+    limit_to=None,
+    **filters,
+) -> Generator[BidsComponent, None, None]:
     """Grabs files using pybids and creates snakemake-friendly lists
 
     Parameters
@@ -806,16 +918,14 @@ def _get_lists_from_bids(
             # a custom path was specified for this input, skip pybids:
             # get input_wildcards by parsing path for {} entries (using a set
             # to get unique only)
-            # get input_zip_lists by using glob_wildcards (but need to modify
+            # get zip_lists by using glob_wildcards (but need to modify
             # to deal with multiple wildcards
 
-            input_path = pybids_inputs[input_name]["custom_path"]
-            input_zip_lists, input_lists, input_wildcards = _parse_custom_path(
-                input_path, **pybids_inputs[input_name]["filters"], **filters
-            )
-            yield _BidsLists(
-                input_name, input_path, input_zip_lists, input_lists, input_wildcards
+            path: str = pybids_inputs[input_name]["custom_path"]
+            zip_lists = _parse_custom_path(
+                path, **pybids_inputs[input_name]["filters"], **filters
             )
+            yield BidsComponent(input_name, path, zip_lists)
             continue
 
         if bids_layout is None:
@@ -826,13 +936,13 @@ def _get_lists_from_bids(
             )
             continue
 
-        input_wildcards = {}
-        input_zip_lists = defaultdict(list)
-        input_lists = defaultdict(set)
+        zip_lists = defaultdict(list)
         paths = set()
-        for img in bids_layout.get(
-            **pybids_inputs[input_name].get("filters", {}), **filters
-        ):
+        pybids_filters = {
+            key: Query.ANY if val is True else Query.NONE if val is False else val
+            for key, val in pybids_inputs[input_name].get("filters", {}).items()
+        }
+        for img in bids_layout.get(**pybids_filters, **filters):
             wildcards: List[str] = [
                 wildcard
                 for wildcard in pybids_inputs[input_name].get("wildcards", [])
@@ -840,37 +950,27 @@ def _get_lists_from_bids(
             ]
             _logger.debug("Wildcards %s found entities for %s", wildcards, img.path)
 
-            input_path, parsed_wildcards = _parse_bids_path(img.path, wildcards)
+            path, parsed_wildcards = _parse_bids_path(img.path, wildcards)
 
             for wildcard_name, value in parsed_wildcards.items():
-                input_zip_lists[wildcard_name].append(value)
-                input_lists[wildcard_name].add(value)
-                input_wildcards[wildcard_name] = f"{{{wildcard_name}}}"
+                zip_lists[wildcard_name].append(value)
 
-            paths.add(input_path)
+            paths.add(path)
 
         # now, check to see if unique
         if len(paths) == 0:
             _logger.warning("No images found for %s", input_name)
             continue
-        if len(paths) > 1:
-            _logger.warning(
-                "More than one snakemake filename for %s, taking the "
-                "first. To correct this, use the --filter_%s option to "
-                "narrow the search. Found filenames: %s",
-                input_name,
-                input_name,
-                paths,
+        try:
+            path = itx.one(paths)
+        except ValueError:
+            raise ConfigError(  # pylint: disable=raise-missing-from
+                f"More than one snakemake filename for {input_name}, taking the "
+                f"first. To correct this, use the --filter_{input_name} option to "
+                f"narrow the search. Found filenames: {paths}"
             )
 
-        input_path = list(paths)[0]
-
-        # convert sets to lists
-        input_lists = {key: list(val) for key, val in input_lists.items()}
-
-        yield _BidsLists(
-            input_name, input_path, input_zip_lists, input_lists, input_wildcards
-        )
+        yield BidsComponent(input_name, path, zip_lists)
 
 
 def get_wildcard_constraints(image_types):
