@@ -11,7 +11,7 @@ def get_parser():
 
 
 def main():
-    app = SnakeBidsApp(Path(__file__).resolve().parent)  # to get repository root
+    app = SnakeBidsApp(Path(__file__).resolve().parent.parent)  # to get repository root
     app.run_snakemake()
 
 
